@@ -31,9 +31,11 @@ from .base import DependencyException, DependencyMethods
 from .base import ExternalDependency, NonExistingExternalProgram
 from .base import ExtraFrameworkDependency, PkgConfigDependency
 from .base import ConfigToolDependency, DependencyFactory
+from .base import find_external_program
 
 if T.TYPE_CHECKING:
     from ..environment import Environment
+    from .base import ExternalProgram
 
 
 class GLDependencySystem(ExternalDependency):
@@ -227,24 +229,21 @@ class QtBaseDependency(ExternalDependency):
         bins = ['moc', 'uic', 'rcc', 'lrelease']
         found = {b: NonExistingExternalProgram(name='{}-{}'.format(b, self.name))
                  for b in bins}
+        wanted = '== {}'.format(self.version)
 
         def gen_bins():
             for b in bins:
                 if self.bindir:
-                    yield os.path.join(self.bindir, b), b, False
-                yield '{}-{}'.format(b, self.name), b, False
-                yield b, b, self.required if b != 'lrelease' else False
+                    yield os.path.join(self.bindir, b), b
+                # prefer the <tool>-qt<version> of the tool to the plain one, as we
+                # don't know what the unsuffixed one points to without calling it.
+                yield '{}-{}'.format(b, self.name), b
+                yield b, b
 
-        for b, name, required in gen_bins():
+        for b, name in gen_bins():
             if found[name].found():
                 continue
 
-            # prefer the <tool>-qt<version> of the tool to the plain one, as we
-            # don't know what the unsuffixed one points to without calling it.
-            p = interp_obj.find_program_impl([b], silent=True, required=required).held_object
-            if not p.found():
-                continue
-
             if name == 'lrelease':
                 arg = ['-version']
             elif mesonlib.version_compare(self.version, '>= 5'):
@@ -253,12 +252,18 @@ class QtBaseDependency(ExternalDependency):
                 arg = ['-v']
 
             # Ensure that the version of qt and each tool are the same
-            _, out, err = mesonlib.Popen_safe(p.get_command() + arg)
-            if b.startswith('lrelease') or not self.version.startswith('4'):
-                care = out
-            else:
-                care = err
-            if mesonlib.version_compare(self.version, '== {}'.format(care.split(' ')[-1])):
+            def get_version(p):
+                _, out, err = mesonlib.Popen_safe(p.get_command() + arg)
+                if b.startswith('lrelease') or not self.version.startswith('4'):
+                    care = out
+                else:
+                    care = err
+                return care.split(' ')[-1].replace(')', '').strip()
+
+            p = interp_obj.find_program_impl([b], required=False,
+                                             version_func=get_version,
+                                             wanted=wanted).held_object
+            if p.found():
                 found[name] = p
 
         return tuple([found[b] for b in bins])
@@ -324,10 +329,9 @@ class QtBaseDependency(ExternalDependency):
             if prefix:
                 self.bindir = os.path.join(prefix, 'bin')
 
-    def search_qmake(self):
+    def search_qmake(self) -> T.Generator['ExternalProgram', None, None]:
         for qmake in ('qmake-' + self.name, 'qmake'):
-            for potential_qmake in self.search_tool(qmake, 'QMake', [qmake]):
-                yield potential_qmake
+            yield from find_external_program(self.env, self.for_machine, qmake, 'QMake', [qmake])
 
     def _qmake_detect(self, mods, kwargs):
         for qmake in self.search_qmake():
@@ -378,9 +382,9 @@ class QtBaseDependency(ExternalDependency):
 
         # Use the buildtype by default, but look at the b_vscrt option if the
         # compiler supports it.
-        is_debug = self.env.coredata.get_builtin_option('buildtype') == 'debug'
-        if 'b_vscrt' in self.env.coredata.base_options:
-            if self.env.coredata.base_options['b_vscrt'].value in ('mdd', 'mtd'):
+        is_debug = self.env.coredata.get_option(mesonlib.OptionKey('buildtype')) == 'debug'
+        if mesonlib.OptionKey('b_vscrt') in self.env.coredata.options:
+            if self.env.coredata.options[mesonlib.OptionKey('b_vscrt')].value in {'mdd', 'mtd'}:
                 is_debug = True
         modules_lib_suffix = self._get_modules_lib_suffix(is_debug)
 
@@ -540,15 +544,28 @@ class Qt5Dependency(QtBaseDependency):
         return _qt_get_private_includes(mod_inc_dir, module, self.version)
 
 
+class Qt6Dependency(QtBaseDependency):
+    def __init__(self, env, kwargs):
+        QtBaseDependency.__init__(self, 'qt6', env, kwargs)
+
+    def get_pkgconfig_host_bins(self, core):
+        return core.get_pkgconfig_variable('host_bins', {})
+
+    def get_private_includes(self, mod_inc_dir, module):
+        return _qt_get_private_includes(mod_inc_dir, module, self.version)
+
+
 class SDL2DependencyConfigTool(ConfigToolDependency):
 
     tools = ['sdl2-config']
     tool_name = 'sdl2-config'
 
-    @staticmethod
-    def finish_init(ctdep):
-        ctdep.compile_args = ctdep.get_config_value(['--cflags'], 'compile_args')
-        ctdep.link_args = ctdep.get_config_value(['--libs'], 'link_args')
+    def __init__(self, name: str, environment: 'Environment', kwargs: T.Dict[str, T.Any]):
+        super().__init__(name, environment, kwargs)
+        if not self.is_found:
+            return
+        self.compile_args = self.get_config_value(['--cflags'], 'compile_args')
+        self.link_args = self.get_config_value(['--libs'], 'link_args')
 
     @staticmethod
     def get_methods():
