@@ -14,11 +14,11 @@
 
 import sys, os
 import subprocess
-import shlex
 import shutil
 import argparse
-from ..mesonlib import MesonException, Popen_safe
+from ..mesonlib import MesonException, Popen_safe, is_windows, is_cygwin, split_args
 from . import destdir_join
+import typing as T
 
 parser = argparse.ArgumentParser()
 
@@ -28,6 +28,7 @@ parser.add_argument('--subdir', dest='subdir')
 parser.add_argument('--headerdirs', dest='headerdirs')
 parser.add_argument('--mainfile', dest='mainfile')
 parser.add_argument('--modulename', dest='modulename')
+parser.add_argument('--moduleversion', dest='moduleversion')
 parser.add_argument('--htmlargs', dest='htmlargs', default='')
 parser.add_argument('--scanargs', dest='scanargs', default='')
 parser.add_argument('--scanobjsargs', dest='scanobjsargs', default='')
@@ -45,28 +46,52 @@ parser.add_argument('--ignore-headers', dest='ignore_headers', default='')
 parser.add_argument('--namespace', dest='namespace', default='')
 parser.add_argument('--mode', dest='mode', default='')
 parser.add_argument('--installdir', dest='install_dir')
+parser.add_argument('--run', dest='run', default='')
+for tool in ['scan', 'scangobj', 'mkdb', 'mkhtml', 'fixxref']:
+    program_name = 'gtkdoc-' + tool
+    parser.add_argument('--' + program_name, dest=program_name.replace('-', '_'))
+
+def gtkdoc_run_check(cmd: T.List[str], cwd: str, library_paths: T.Optional[T.List[str]] = None) -> None:
+    if library_paths is None:
+        library_paths = []
 
-def gtkdoc_run_check(cmd, cwd, library_path=None):
     env = dict(os.environ)
-    if library_path:
-        env['LD_LIBRARY_PATH'] = library_path
+    if is_windows() or is_cygwin():
+        if 'PATH' in env:
+            library_paths.extend(env['PATH'].split(os.pathsep))
+        env['PATH'] = os.pathsep.join(library_paths)
+    else:
+        if 'LD_LIBRARY_PATH' in env:
+            library_paths.extend(env['LD_LIBRARY_PATH'].split(os.pathsep))
+        env['LD_LIBRARY_PATH'] = os.pathsep.join(library_paths)
+
+    if is_windows():
+        cmd.insert(0, sys.executable)
+
     # Put stderr into stdout since we want to print it out anyway.
     # This preserves the order of messages.
     p, out = Popen_safe(cmd, cwd=cwd, env=env, stderr=subprocess.STDOUT)[0:2]
     if p.returncode != 0:
-        err_msg = ["{!r} failed with status {:d}".format(cmd[0], p.returncode)]
+        err_msg = ["{!r} failed with status {:d}".format(cmd, p.returncode)]
         if out:
             err_msg.append(out)
         raise MesonException('\n'.join(err_msg))
     elif out:
-        print(out)
+        # Unfortunately Windows cmd.exe consoles may be using a codepage
+        # that might choke print() with a UnicodeEncodeError, so let's
+        # ignore such errors for now, as a compromise as we are outputting
+        # console output here...
+        try:
+            print(out)
+        except UnicodeEncodeError:
+            pass
 
-def build_gtkdoc(source_root, build_root, doc_subdir, src_subdirs,
-                 main_file, module,
-                 html_args, scan_args, fixxref_args, mkdb_args,
-                 gobject_typesfile, scanobjs_args, ld, cc, ldflags, cflags,
-                 html_assets, content_files, ignore_headers, namespace,
-                 expand_content_files, mode):
+def build_gtkdoc(source_root: str, build_root: str, doc_subdir: str, src_subdirs: T.List[str],
+                 main_file: str, module: str, module_version: str,
+                 html_args: T.List[str], scan_args: T.List[str], fixxref_args: T.List[str], mkdb_args: T.List[str],
+                 gobject_typesfile: str, scanobjs_args: T.List[str], run: str, ld: str, cc: str, ldflags: str, cflags: str,
+                 html_assets: T.List[str], content_files: T.List[str], ignore_headers: T.List[str], namespace: str,
+                 expand_content_files: T.List[str], mode: str, options: argparse.Namespace) -> None:
     print("Building documentation for %s" % module)
 
     src_dir_args = []
@@ -93,9 +118,12 @@ def build_gtkdoc(source_root, build_root, doc_subdir, src_subdirs,
 
     # Copy files to build directory
     for f in content_files:
-        f_abs = os.path.join(doc_src, f)
-        shutil.copyfile(f_abs, os.path.join(
-            abs_out, os.path.basename(f_abs)))
+        # FIXME: Use mesonlib.File objects so we don't need to do this
+        if not os.path.isabs(f):
+            f = os.path.join(doc_src, f)
+        elif os.path.commonpath([f, build_root]) == build_root:
+            continue
+        shutil.copyfile(f, os.path.join(abs_out, os.path.basename(f)))
 
     shutil.rmtree(htmldir, ignore_errors=True)
     try:
@@ -107,29 +135,35 @@ def build_gtkdoc(source_root, build_root, doc_subdir, src_subdirs,
         f_abs = os.path.join(doc_src, f)
         shutil.copyfile(f_abs, os.path.join(htmldir, os.path.basename(f_abs)))
 
-    scan_cmd = ['gtkdoc-scan', '--module=' + module] + src_dir_args
+    scan_cmd = [options.gtkdoc_scan, '--module=' + module] + src_dir_args
     if ignore_headers:
         scan_cmd.append('--ignore-headers=' + ' '.join(ignore_headers))
     # Add user-specified arguments
     scan_cmd += scan_args
     gtkdoc_run_check(scan_cmd, abs_out)
 
+    # Use the generated types file when available, otherwise gobject_typesfile
+    # would often be a path to source dir instead of build dir.
+    if '--rebuild-types' in scan_args:
+        gobject_typesfile = os.path.join(abs_out, module + '.types')
+
     if gobject_typesfile:
-        scanobjs_cmd = ['gtkdoc-scangobj'] + scanobjs_args + ['--types=' + gobject_typesfile,
-                                                              '--module=' + module,
-                                                              '--cflags=' + cflags,
-                                                              '--ldflags=' + ldflags,
-                                                              '--ld=' + ld]
+        scanobjs_cmd = [options.gtkdoc_scangobj] + scanobjs_args
+        scanobjs_cmd += ['--types=' + gobject_typesfile,
+                         '--module=' + module,
+                         '--run=' + run,
+                         '--cflags=' + cflags,
+                         '--ldflags=' + ldflags,
+                         '--cc=' + cc,
+                         '--ld=' + ld,
+                         '--output-dir=' + abs_out]
 
         library_paths = []
-        for ldflag in shlex.split(ldflags):
+        for ldflag in split_args(ldflags):
             if ldflag.startswith('-Wl,-rpath,'):
                 library_paths.append(ldflag[11:])
-        if 'LD_LIBRARY_PATH' in os.environ:
-            library_paths.append(os.environ['LD_LIBRARY_PATH'])
-        library_path = ':'.join(library_paths)
 
-        gtkdoc_run_check(scanobjs_cmd, abs_out, library_path)
+        gtkdoc_run_check(scanobjs_cmd, build_root, library_paths)
 
     # Make docbook files
     if mode == 'auto':
@@ -146,7 +180,7 @@ def build_gtkdoc(source_root, build_root, doc_subdir, src_subdirs,
     else: # none
         modeflag = None
 
-    mkdb_cmd = ['gtkdoc-mkdb',
+    mkdb_cmd = [options.gtkdoc_mkdb,
                 '--module=' + module,
                 '--output-format=xml',
                 '--expand-content-files=' + ' '.join(expand_content_files),
@@ -155,7 +189,7 @@ def build_gtkdoc(source_root, build_root, doc_subdir, src_subdirs,
         mkdb_cmd.append('--name-space=' + namespace)
     if modeflag:
         mkdb_cmd.append(modeflag)
-    if len(main_file) > 0:
+    if main_file:
         # Yes, this is the flag even if the file is in xml.
         mkdb_cmd.append('--main-sgml-file=' + main_file)
     # Add user-specified arguments
@@ -163,48 +197,52 @@ def build_gtkdoc(source_root, build_root, doc_subdir, src_subdirs,
     gtkdoc_run_check(mkdb_cmd, abs_out)
 
     # Make HTML documentation
-    mkhtml_cmd = ['gtkdoc-mkhtml',
+    mkhtml_cmd = [options.gtkdoc_mkhtml,
                   '--path=' + ':'.join((doc_src, abs_out)),
                   module,
                   ] + html_args
-    if len(main_file) > 0:
+    if main_file:
         mkhtml_cmd.append('../' + main_file)
     else:
         mkhtml_cmd.append('%s-docs.xml' % module)
     # html gen must be run in the HTML dir
-    gtkdoc_run_check(mkhtml_cmd, os.path.join(abs_out, 'html'))
+    gtkdoc_run_check(mkhtml_cmd, htmldir)
 
     # Fix cross-references in HTML files
-    fixref_cmd = ['gtkdoc-fixxref',
+    fixref_cmd = [options.gtkdoc_fixxref,
                   '--module=' + module,
                   '--module-dir=html'] + fixxref_args
     gtkdoc_run_check(fixref_cmd, abs_out)
 
-def install_gtkdoc(build_root, doc_subdir, install_prefix, datadir, module):
+    if module_version:
+        shutil.move(os.path.join(htmldir, '{}.devhelp2'.format(module)),
+                    os.path.join(htmldir, '{}-{}.devhelp2'.format(module, module_version)))
+
+def install_gtkdoc(build_root: str, doc_subdir: str, install_prefix: str, datadir: str, module: str) -> None:
     source = os.path.join(build_root, doc_subdir, 'html')
     final_destination = os.path.join(install_prefix, datadir, module)
     shutil.rmtree(final_destination, ignore_errors=True)
     shutil.copytree(source, final_destination)
 
-def run(args):
+def run(args: T.List[str]) -> int:
     options = parser.parse_args(args)
-    if len(options.htmlargs) > 0:
+    if options.htmlargs:
         htmlargs = options.htmlargs.split('@@')
     else:
         htmlargs = []
-    if len(options.scanargs) > 0:
+    if options.scanargs:
         scanargs = options.scanargs.split('@@')
     else:
         scanargs = []
-    if len(options.scanobjsargs) > 0:
+    if options.scanobjsargs:
         scanobjsargs = options.scanobjsargs.split('@@')
     else:
         scanobjsargs = []
-    if len(options.fixxrefargs) > 0:
+    if options.fixxrefargs:
         fixxrefargs = options.fixxrefargs.split('@@')
     else:
         fixxrefargs = []
-    if len(options.mkdbargs) > 0:
+    if options.mkdbargs:
         mkdbargs = options.mkdbargs.split('@@')
     else:
         mkdbargs = []
@@ -215,12 +253,14 @@ def run(args):
         options.headerdirs.split('@@'),
         options.mainfile,
         options.modulename,
+        options.moduleversion,
         htmlargs,
         scanargs,
         fixxrefargs,
         mkdbargs,
         options.gobject_typesfile,
         scanobjsargs,
+        options.run,
         options.ld,
         options.cc,
         options.ldflags,
@@ -230,12 +270,18 @@ def run(args):
         options.ignore_headers.split('@@') if options.ignore_headers else [],
         options.namespace,
         options.expand_content_files.split('@@') if options.expand_content_files else [],
-        options.mode)
+        options.mode,
+        options)
 
     if 'MESON_INSTALL_PREFIX' in os.environ:
         destdir = os.environ.get('DESTDIR', '')
         install_prefix = destdir_join(destdir, os.environ['MESON_INSTALL_PREFIX'])
-        install_dir = options.install_dir if options.install_dir else options.modulename
+        if options.install_dir:
+            install_dir = options.install_dir
+        else:
+            install_dir = options.modulename
+            if options.moduleversion:
+                install_dir += '-' + options.moduleversion
         if os.path.isabs(install_dir):
             install_dir = destdir_join(destdir, install_dir)
         install_gtkdoc(options.builddir,
